#include "hmbdc/Copyright.hpp"
#pragma once

#include "hmbdc/app/Client.hpp"
#include "hmbdc/text/LfbStream.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/pattern/GuardedSingleton.hpp"
#include  <mutex>

#ifdef HMBDC_LOG_CONTEXT //if defined, it has to be 'compile time sized' Context
#include "hmbdc/app/Context.hpp"
#include <iostream>
#include <string>
#include <memory>
#include <tuple>

#define HMBDC_LOG_D(...)   hmbdc::app::LoggerT<HMBDC_LOG_CONTEXT>::instance().LOG_D(__VA_ARGS__, LogTrailer(__FILE__, __LINE__))
#define HMBDC_LOG_N(...)   hmbdc::app::LoggerT<HMBDC_LOG_CONTEXT>::instance().LOG_N(__VA_ARGS__, LogTrailer(__FILE__, __LINE__))
#define HMBDC_LOG_W(...)   hmbdc::app::LoggerT<HMBDC_LOG_CONTEXT>::instance().LOG_W(__VA_ARGS__, LogTrailer(__FILE__, __LINE__))
#define HMBDC_LOG_C(...)   hmbdc::app::LoggerT<HMBDC_LOG_CONTEXT>::instance().LOG_C(__VA_ARGS__, LogTrailer(__FILE__, __LINE__))
#define HMBDC_LOG_DEBUG(x) hmbdc::app::LoggerT<HMBDC_LOG_CONTEXT>::instance().LOG_D(#x "=", x,   LogTrailer(__FILE__, __LINE__))

#define HMBDC_LOG_d(...)   hmbdc::app::LoggerT<HMBDC_LOG_CONTEXT>::instance().LOG_D(__VA_ARGS__, EmptyLogTrailer())
#define HMBDC_LOG_n(...)   hmbdc::app::LoggerT<HMBDC_LOG_CONTEXT>::instance().LOG_N(__VA_ARGS__, EmptyLogTrailer())
#define HMBDC_LOG_w(...)   hmbdc::app::LoggerT<HMBDC_LOG_CONTEXT>::instance().LOG_W(__VA_ARGS__, EmptyLogTrailer())
#define HMBDC_LOG_c(...)   hmbdc::app::LoggerT<HMBDC_LOG_CONTEXT>::instance().LOG_C(__VA_ARGS__, EmptyLogTrailer())

#define HMBDC_LOGGER hmbdc::app::LoggerT<HMBDC_LOG_CONTEXT>

#else

#define HMBDC_LOG_D(...)   if (&hmbdc::app::SimpleLogger::instance()) hmbdc::app::SimpleLogger::instance().LOG_D(hmbdc::time::SysTime::now(), g_SimpleLogLevelStr[0], __VA_ARGS__, LogTrailer(__FILE__, __LINE__))
#define HMBDC_LOG_N(...)   if (&hmbdc::app::SimpleLogger::instance()) hmbdc::app::SimpleLogger::instance().LOG_N(hmbdc::time::SysTime::now(), g_SimpleLogLevelStr[1], __VA_ARGS__, LogTrailer(__FILE__, __LINE__))
#define HMBDC_LOG_W(...)   if (&hmbdc::app::SimpleLogger::instance()) hmbdc::app::SimpleLogger::instance().LOG_W(hmbdc::time::SysTime::now(), g_SimpleLogLevelStr[2], __VA_ARGS__, LogTrailer(__FILE__, __LINE__))
#define HMBDC_LOG_C(...)   if (&hmbdc::app::SimpleLogger::instance()) hmbdc::app::SimpleLogger::instance().LOG_C(hmbdc::time::SysTime::now(), g_SimpleLogLevelStr[3], __VA_ARGS__, LogTrailer(__FILE__, __LINE__))
#define HMBDC_LOG_DEBUG(x) if (&hmbdc::app::SimpleLogger::instance()) hmbdc::app::SimpleLogger::instance().LOG_D(hmbdc::time::SysTime::now(), g_SimpleLogLevelStr[0], #x "=", x,   LogTrailer(__FILE__, __LINE__))
#define HMBDC_LOG_d(...)   if (&hmbdc::app::SimpleLogger::instance()) hmbdc::app::SimpleLogger::instance().LOG_D(hmbdc::time::SysTime::now(), g_SimpleLogLevelStr[0], __VA_ARGS__, '\n')
#define HMBDC_LOG_n(...)   if (&hmbdc::app::SimpleLogger::instance()) hmbdc::app::SimpleLogger::instance().LOG_N(hmbdc::time::SysTime::now(), g_SimpleLogLevelStr[1], __VA_ARGS__, '\n')
#define HMBDC_LOG_w(...)   if (&hmbdc::app::SimpleLogger::instance()) hmbdc::app::SimpleLogger::instance().LOG_W(hmbdc::time::SysTime::now(), g_SimpleLogLevelStr[2], __VA_ARGS__, '\n')
#define HMBDC_LOG_c(...)   if (&hmbdc::app::SimpleLogger::instance()) hmbdc::app::SimpleLogger::instance().LOG_C(hmbdc::time::SysTime::now(), g_SimpleLogLevelStr[3], __VA_ARGS__, '\n')

#define HMBDC_LOGGER hmbdc::app::SimpleLogger
#endif //HMBDC_LOG_CONTEXT

namespace hmbdc { namespace app {

char const g_LogLevelStr[][12] = {
    "DEBUG :    ", 
    "NOTICE:    ", 
    "WARNING:   ", 
    "CRITICAL:  "
};

char const g_SimpleLogLevelStr[][12 + 1] = {
    " debug :    ", 
    " notice:    ", 
    " warning:   ", 
    " critical:  "
};

struct LogTrailer {
	LogTrailer(char const* const file,  int line)
	: f(file)
	, l(line) {
	}
    char const* const f;
	int l;

	friend std::ostream& operator << (std::ostream& os, LogTrailer const& t) {
		os << ' ' << t.f << ':' << t.l << std::endl;
		return os;
	}
};
struct EmptyLogTrailer {
	friend std::ostream& operator << (std::ostream& os, EmptyLogTrailer const&) {
		os << std::endl;
		return os;
	}
};

/**
 * @class LoggerT<>
 * @brief a high performance async logger that doesn't penalize logging threads as much
 * when the logging load is heavy
 * @details see @example hmbdc-log.cpp
 * 
 */
template <typename Ctx>
struct LoggerT
: Client<LoggerT<Ctx>, LoggingT<Ctx::MAX_MESSAGE_SIZE>>
, text::lfb_stream::OStringStream<typename Ctx::Buffer
    , MessageWrap<LoggingT<Ctx::MAX_MESSAGE_SIZE>>
    , LoggingT<Ctx::MAX_MESSAGE_SIZE>::typeTag
  >
, pattern::GuardedSingleton<LoggerT<Ctx>> {
	static_assert(Ctx::MAX_MESSAGE_SIZE != 0
		, "HMBDC_LOG_CONTEXT needs to be 'compile time sized'");
	using Logging = LoggingT<Ctx::MAX_MESSAGE_SIZE>;
	friend struct pattern::SingletonGuardian<LoggerT<Ctx>>;

	enum Level {
		L_DEBUG = 0,
		L_NOTICE,
		L_WARNING,
		L_CRITICAL,
		L_OFF
	};
private:
	std::ostream& log_;
	static std::unique_ptr<Ctx> pCtx_s; //Logger is a singleton
	Ctx& ctx_;
	Level minLevel_;
	std::string schedPolicy_;
	int priority_;

	using Stream = text::lfb_stream::OStringStream<typename Ctx::Buffer
	    , MessageWrap<Logging>, Logging::typeTag>;

	LoggerT(std::ostream& log
		, uint64_t cpuAffinityMask = 0
		, char const* schedPolicy = "SCHED_IDLE"
		, int priority = 0
		, uint16_t logBufferSizePower2Num = 8)
	: Stream((pCtx_s = std::unique_ptr<Ctx>(new Ctx(logBufferSizePower2Num)))->buffer())
	, log_(log)
	, ctx_(*pCtx_s)
	, minLevel_(L_DEBUG)
	, schedPolicy_(schedPolicy)
	, priority_(priority) {
#ifdef HMBDC_LOG_CONTEXT		
		ctx_.start(*this
			, cpuAffinityMask?cpuAffinityMask:(1ul << std::thread::hardware_concurrency()) - 1ul);
#endif		
	}

	~LoggerT() {
		if (pCtx_s) {
			pCtx_s->stop();
			pCtx_s->join();
		}
	}

	LoggerT(std::ostream& log, Ctx& ctx
		, char const* schedPolicy = "SCHED_IDLE"
		, int priority = 0)
	: Stream(ctx.buffer())
	, log_(log)
	, ctx_(ctx)
	, minLevel_(L_DEBUG)
	, schedPolicy_(schedPolicy)
	, priority_(priority) {
	}

	struct LogHeader {
		LogHeader(Level level)
		: l(level)
		, ts(hmbdc::time::SysTime::now()) {
		}

		Level l;
		hmbdc::time::SysTime ts;

		friend std::ostream& operator << (std::ostream& os, LogHeader const& h) {
			os << h.ts << " " << g_LogLevelStr[h.l];
			return os;
		}
	};

public:
    char const* hmbdcName() const { return "logger"; }
    void setMinLogLevel(Level minLevel) {
    	minLevel_ = minLevel;
    }

    std::tuple<char const*, int> schedSpec() const {
        return std::make_tuple(schedPolicy_.c_str(), priority_);
    }

	template <typename ...Args>
	void LOG_D(Args&&... args) {
#ifndef NDEBUG
		Stream::operator()(LogHeader(L_DEBUG), std::forward<Args>(args)...);
#endif
	}

	template <typename ...Args>
	void LOG_N(Args&&... args) {
		if (minLevel_ <= L_NOTICE)
			Stream::operator()(LogHeader(L_NOTICE), std::forward<Args>(args)...);
	}
	template <typename ...Args>
	void LOG_W(Args&&... args) {
		if (minLevel_ <= L_WARNING)
		    Stream::operator()(LogHeader(L_WARNING), std::forward<Args>(args)...);
	}
	template <typename ...Args>
	void LOG_C(Args&&... args) {
		if (minLevel_ <= L_CRITICAL)
			Stream::operator()(LogHeader(L_CRITICAL), std::forward<Args>(args)...);
	}

	void handleMessageCb(Logging& logItem) {
		Stream::dump(log_, logItem);
	}
};

template <typename Ctx>
std::unique_ptr<Ctx> LoggerT<Ctx>::pCtx_s;

/**
 * @brief a very straightforward logger that works safely
 * @details when HMBDC_LOG_CONTEXT is undefined, you get functionality of this
 * 
 */
struct SimpleLogger 
: pattern::GuardedSingleton<SimpleLogger> {
	friend struct pattern::SingletonGuardian<SimpleLogger>;

	enum Level {
		L_DEBUG = 0,
		L_NOTICE,
		L_WARNING,
		L_CRITICAL,
		L_OFF
	};
	template <typename ...Args>
	void LOG_D(Args&&... args) {
#ifndef NDEBUG
		     std::lock_guard<std::recursive_mutex> g(mutex_);
		     log(std::forward<Args>(args)...);
#endif  
	}

	template <typename ...Args>
	void LOG_N(Args&&... args) {
		if (minLevel_ <= L_NOTICE) {
		    std::lock_guard<std::recursive_mutex> g(mutex_);
			log(std::forward<Args>(args)...);
		}
	} 
	template <typename ...Args>
	void LOG_W(Args&&... args) {
		if (minLevel_ <= L_WARNING) {
		    std::lock_guard<std::recursive_mutex> g(mutex_);
			log(std::forward<Args>(args)...);
		}
	}
	template <typename ...Args>
	void LOG_C(Args&&... args) {
		if (minLevel_ <= L_CRITICAL) {
		    std::lock_guard<std::recursive_mutex> g(mutex_);
			log(std::forward<Args>(args)...);
		}
	}

private:
	template <typename Arg, typename ...Args>
	void log(Arg&& arg, Args&&... args) {
		log_ << std::forward<Arg>(arg);
		log(std::forward<Args>(args)...);
	}
	void log() {
	}
	template <typename ... NoOpArgs>
	SimpleLogger(std::ostream& log, NoOpArgs&&...)
	: log_(log)
	, minLevel_(L_DEBUG) 
	{}
    std::ostream& log_;
	Level minLevel_;
    std::recursive_mutex mutex_;
};
}}

