#include "hmbdc/Copyright.hpp"
#pragma once 

#include "hmbdc/comm/Topic.hpp"
#include "hmbdc/app/Message.hpp"

#include <type_traits>
#include <ostream>

namespace hmbdc { namespace app { namespace netmap {

struct TransportMessageHeader {
    uint8_t topicLen;
    uint16_t& messagePayloadLen() {
        return *reinterpret_cast<uint16_t*>(payload());
    }
    uint16_t const& messagePayloadLen()  const {
        return *reinterpret_cast<uint16_t const*>(payload());
    }

//copy message into the right place in memory
    template <typename M>
    static TransportMessageHeader*
    copyTo(void* addrIn, comm::Topic const& t, M&& m) {
        using Message = typename std::remove_reference<M>::type;
        auto addr = (char*)addrIn;
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        new (addr + sizeof(TransportMessageHeader) + tl) 
            MessageWrap<Message>(std::forward<M>(m));
        h->messagePayloadLen() = sizeof(MessageWrap<Message>);
        return h;
    }

    static TransportMessageHeader*
    copyTo(void* addrIn, comm::Topic const& t, uint16_t tag, void const* bytes, size_t len) {
        auto addr = (char*)addrIn;
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<JustBytes>(tag, bytes, len);
        h->messagePayloadLen() = sizeof(MessageWrap<JustBytes>) - sizeof(MessageWrap<JustBytes>::payload) + len;
        return h;
    }

    template <typename Message, typename ... Args>
    static TransportMessageHeader*
    copyToInPlace(void* addrIn, comm::Topic const& t, Args&&... args) {
        char* addr = (char*)addrIn;
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<Message>(std::forward<Args>(args)...);
        h->messagePayloadLen() = sizeof(MessageWrap<Message>);
        return h;
    }

    std::pair<char const*, char const*> topic() const {
        char const* b = reinterpret_cast<const char*>(this) 
            + sizeof(TransportMessageHeader); 
        return std::make_pair(b, b + topicLen);
    }

    void const* payload() const {
        return reinterpret_cast<const char*>(this) 
            + sizeof(TransportMessageHeader) + topicLen; 
    }

    void * payload() {
        return reinterpret_cast<char*>(this) 
            + sizeof(TransportMessageHeader) + topicLen; 
    }

    uint16_t typeTag() const {
        auto h = static_cast<app::MessageHead const*>(payload());
        return h->typeTag;
    }

    template <typename Message>
    Message& wrapped() {
        auto wrap = static_cast<app::MessageWrap<Message>*>(payload());
        return wrap->payload;
    }

    template <typename Message>
    Message const& wrapped() const {
        auto wrap = static_cast<app::MessageWrap<Message> const *>(payload());
        return wrap->payload;
    }

    size_t wireSize() const {
        return sizeof(TransportMessageHeader) + topicLen 
            + messagePayloadLen();
    }    

    size_t wireSizeContainsTopic() const {
        return sizeof(TransportMessageHeader) + topicLen;
    }

    template <typename Message>
    static 
    size_t wireSize(comm::Topic const& t) {

        return sizeof(TransportMessageHeader) + strnlen(t.c_str(), sizeof(t)) 
            + sizeof(MessageWrap<Message>);
    }
    
} __attribute__((packed));

struct Subscribe
: hasTag<301> {
    Subscribe(comm::Topic const&topic)
    : topic(topic){}
    comm::Topic topic;
};

struct Unsubscribe 
: hasTag<302> {
    Unsubscribe(comm::Topic const&topic)
    : topic(topic){}
    comm::Topic topic;
};

}}}
