#include "hmbdc/Copyright.hpp"
#pragma once

#include "hmbdc/app/Base.hpp"
#include "hmbdc/text/StringTrieSet.hpp"
#include "hmbdc/comm/Topic.hpp"
#include "hmbdc/comm/inet/Misc.hpp"
#include "hmbdc/time/Timers.hpp"
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"

#include <boost/regex.hpp>
#include <boost/asio.hpp>

#include <memory>

extern thread_local std::shared_ptr<boost::asio::io_service> gtIos;

namespace hmbdc { namespace app { namespace tcpcast {

extern thread_local std::shared_ptr<boost::asio::io_service> gtIos;

struct Transport {
    using ptr = std::shared_ptr<Transport>;

    Transport(Config const& cfg)
    : config_(cfg)
    , pIos_(nullptr)
    , tcpcastAdTopic_("tcpcastad") {
        cfg (hmbdcName_, "hmbdcName")
            (schedPolicy_, "schedPolicy")
            (schedPriority_, "schedPriority")
        ;
    }

    void initInThread() {
        if (!::gtIos) {
            ::gtIos.reset(new boost::asio::io_service(1u));
        }
        pIos_ = ::gtIos;
    }

    bool operator == (Transport const& other ) const {
        return &config_ == &other.config_;
    }

    bool operator < (Transport const& other ) const {
        return &config_ < &other.config_;
    }

    virtual ~Transport(){}
    
protected:
    Config const config_;
    std::string hmbdcName_;
    std::string schedPolicy_;
    int schedPriority_;
    boost::asio::ip::udp::endpoint endpoint_;    
    std::shared_ptr<boost::asio::io_service> pIos_;
    Topic const tcpcastAdTopic_;
};
}}}
